/*----------------------

   GATE - Geant4 Application for Tomographic Emission 
   OpenGATE Collaboration 
     
   Giovanni Santin <giovanni.santin@cern.ch> 
   Daniel Strul <daniel.strul@iphe.unil.ch> 
     
   Copyright (C) 2002,2003 UNIL/IPHE, CH-1015 Lausanne 

This software is distributed under the terms 
of the GNU Lesser General  Public Licence (LGPL) 
See GATE/LICENSE.txt for further details 
----------------------*/

/*!
  \file GateTrajectoryNavigator.cc
  
  $Log: GateTrajectoryNavigator.cc,v $
  Revision 1.1.1.1  2005/02/03 13:09:53  jdbeen
  gate cluster code for G4.6.2

  Revision 1.8  2003/06/24 16:34:19  gsantin
  reduced warnings

  Revision 1.7  2003/06/20 13:08:25  gsantin
  distance for annih. vertex reconstr. reduced from 1E-6 to 1E-7mm to avoid 3/4 photon vertices

  Revision 1.6  2002/11/25 15:16:28  gsantin
  adaptations for SPECT, handling of photon trajectories not only in pairs

  Revision 1.5  2002/08/11 15:33:25  dstrul
  Cosmetic cleanup: standardized file comments for cleaner doxygen output


  \brief Class GateTrajectoryNavigator
  \brief By Giovanni.Santin@cern.ch
  \brief $Id: GateTrajectoryNavigator.cc,v 1.1.1.1 2005/02/03 13:09:53 jdbeen Exp $
*/
#include "GateTrajectoryNavigator.hh"

#include "G4TrajectoryContainer.hh"
#include "G4Trajectory.hh"

#include "G4Navigator.hh"
#include "G4TransportationManager.hh"

//dk
#include "G4TrajectoryPoint.hh"

G4int GateTrajectoryNavigator::FindPrimType(G4int trackID,G4int gam1_ID, G4int gam2_ID)
{
  // identifies the gamma according to the following classification
  // type 0 = any other
  // type 1 = annihilation      : parent_id==positron_id&&vertex
  // type 2 = direct with e+    : parent->PDGEncoding()==0 
  // type 3 = direct without e+ : parent->PDGEncoding()==0
  // type 4 = secondary positron annihilation (gamma-> e+e- -> 2 gamma + e- e.g. Y86)
  // m_positronTrackID must be known when this routine is called
  // we also set m_PrimEnergy
  G4int PrimType = 0;
  
  if (!m_trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::FindPrimType : ERROR : NULL trajectoryContainer" << G4endl;
  } else {
    // often TrackID is an e- kicked by the gamma inside the crystal
    // we go up and up starting from the trackID, via the parentID's, until
    // the track we find is a primary (its parentID==0) or it comes from the positron
    // or  we find one of the annihilation  gammas -> type 1
    // or it is the ion -> type 2 (positron in event) or 3 (no positron in event)
    G4int lookID = trackID;
    G4bool loop=true;
    G4int n_trajectories = m_trajectoryContainer->entries();
    G4Trajectory* lasttrj = 0;
    G4int parentID = -2; //for parentID!=m_positronTrackID
    PrimType=-1;
//G4cout<<"\n-------------->>> "<<trackID<<G4endl;
    do {
       G4bool found = false;
       loop  = false;
       m_PrimEnergy = 0;
       for (G4int iTrj=0; (iTrj<n_trajectories) && (!found); iTrj++) {
          G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
          if ( lookID == trj->GetTrackID()) {
              found = true;
              G4int PDG = trj->GetPDGEncoding();
              //non gamma is allowed hereif(PDG!=22&&parentID==-1) return 0;//called for a non gamma
//G4cout<<"-------------->>> "<<PDG<<" "<<lookID<<" "<<parentID<<" "<<gam1_ID<<" "<<m_positronTrackID<<G4endl;
              parentID = trj->GetParentID();				
//G4cout<<"-------------->>>  "<<parentID<<G4endl;
             //lookid!=0
            if(lookID==gam1_ID||lookID==gam2_ID){//annihilation
               PrimType = 1;
               m_PrimEnergy=trj->GetInitialMomentum().mag();
            } else if (PDG==0){
              if (m_positronTrackID != 0) PrimType = 2;
              else                        PrimType = 3;
              if(lasttrj!=0) m_PrimEnergy=lasttrj->GetInitialMomentum().mag();
            } else if (parentID!=0&&parentID!=m_positronTrackID) {
              lookID = parentID;
              lasttrj=trj;
              loop = true;
            } else PrimType = 0;
          }
       }		
    } while (loop);
//G4cout<<"-------------->>> "<<PrimType<<G4endl;
}
    return PrimType;
}

//dk end

GateTrajectoryNavigator::GateTrajectoryNavigator()
{
}

GateTrajectoryNavigator::~GateTrajectoryNavigator()
{
}

G4ThreeVector GateTrajectoryNavigator::FindSourcePosition()
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::FindSourcePosition" << G4endl;

  G4ThreeVector sourcePosition = G4ThreeVector();

  G4int sourceIndex = FindSourceIndex();

  if ((sourceIndex < 0) || (sourceIndex >= m_trajectoryContainer->entries())) {
    G4cout << "GateTrajectoryNavigator::FindSourcePosition : WARNING : sourceIndex out of range: " << sourceIndex << G4endl;
  } else {
    G4Trajectory* trjSource = (G4Trajectory*)((*m_trajectoryContainer)[sourceIndex]);
    sourcePosition = ((G4TrajectoryPoint*)(trjSource->GetPoint(0)))->GetPosition();
  }

  return sourcePosition;
}

G4int GateTrajectoryNavigator::FindSourceIndex()
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::FindSourceIndex" << G4endl;
  G4int sourceIndex = -1;
  G4bool found = false;
  G4int n_trajectories = m_trajectoryContainer->entries();
  for (G4int iTrj=0; (iTrj<n_trajectories) && (!found); iTrj++) {
    G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
    if (trj->GetTrackID() == 1) { 
      sourceIndex = iTrj;
      found = true;
    }
  }

  return sourceIndex; 
}


void GateTrajectoryNavigator::SetIonID()
{
  G4int n_trajectories = m_trajectoryContainer->entries();
  for (G4int iTrj=0; iTrj<n_trajectories; iTrj++) {
    G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
    if (trj->GetTrackID() == 1) {
//       G4cout << "GateTrajectoryNavigator::SetIonID: PDG " << trj->GetPDGEncoding() 
// 	     << " Name <" << trj->GetParticleName() << ">" << G4endl;
      if (trj->GetPDGEncoding() == 0) { // 0 == Ion -----> to be changed !! 
	m_ionID = 1;
      } else {
	m_ionID = 0;
      }
    }
  }
}

G4int GateTrajectoryNavigator::FindPositronTrackID()
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::FindPositronTrackID" << G4endl;
  m_positronTrackID = 0;

  SetIonID();
  //  m_ionID = 1;

  if (!m_trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::FindPositronTrackID : ERROR : NULL trajectoryContainer" << G4endl;
  } else {
    G4int n_trajectories = m_trajectoryContainer->entries();
    G4bool found = false;
    for (G4int iTrj=0; (iTrj<n_trajectories) && (!found); iTrj++) {
      G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
      if ((trj->GetParentID() == m_ionID)&&(trj->GetPDGEncoding() == -11)) { // -11 == Positron
	m_positronTrackID = trj->GetTrackID();
	m_positronTrj = trj;
	found = true;
      }
    }
  }
  return m_positronTrackID;
}



std::vector<G4int> GateTrajectoryNavigator::FindAnnihilationGammasTrackID()
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::FindAnnihilationGammasTrackID" << G4endl;
  std::vector<G4int> photonIndices;
  photonIndices.clear();

  if (!m_trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::FindAnnihilationGammasTrackID : ERROR : NULL trajectoryContainer" << G4endl;
  } else {

    // prepare the list of gammas for later analysis
    G4int n_trajectories = m_trajectoryContainer->entries();

    for (G4int iTrj=0; iTrj<n_trajectories; iTrj++) {
      G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
      if (m_positronTrackID != 0) {
	// in case the positronTrackID has been found, wee put in the list only 
	// the gammas generated by the positron
	if ((trj->GetParentID() == m_positronTrackID)
	    && (trj->GetPDGEncoding() == 22) ) { // 22 == Gamma
	  photonIndices.push_back(iTrj);
	}
      } else {
	// in case the positron has not been found, we accept all the photons 
	// either coming from the ion (assuming m_ionID==1) or shooted as primary 
	// (both single and back-to-back pair)
	if (((trj->GetParentID() == 0) || (trj->GetParentID() == m_ionID))
	    && (trj->GetPDGEncoding() == 22) ) { // 22 == Gamma
	  photonIndices.push_back(iTrj);
	}
      }
    }

    // we start the analysis of the gammas in the list

    // in both cases (in case a positron has been found or not) we look for 
    // coincident vertices (annihilation gammas, or gammas from ion decay, or 
    // user defined multiple gamma sources)
    G4int nPh = photonIndices.size();
    if (nPh == 1) {
      // if only 1 gamma, we take it
      G4int i1 = photonIndices[0];
      G4Trajectory* trj1 = (G4Trajectory*)((*m_trajectoryContainer)[i1]);
      m_photonIDVec.push_back(trj1->GetTrackID());
    } else if (nPh >= 2) {
      // if more than 1 gamma, we select those coming from a common vertex
      // this design should be open to more than 2 gammas, even if for the moment not 
      // considered by GateAnalysis
      G4int i1;
      G4int i2;
      for (G4int j1=0; j1<nPh; j1++) {
	for (G4int j2=j1+1; j2<nPh; j2++) {
	  i1 = photonIndices[j1];
	  i2 = photonIndices[j2];
	  if ((i1 >= 0) && (i2 >= 0)) {
	    // both gammas were not already taken
	    G4Trajectory* trj1 = (G4Trajectory*)((*m_trajectoryContainer)[i1]);
	    G4Trajectory* trj2 = (G4Trajectory*)((*m_trajectoryContainer)[i2]);
	    G4ThreeVector vert1 = ((G4TrajectoryPoint*)(trj1->GetPoint(0)))->GetPosition();
	    G4ThreeVector vert2 = ((G4TrajectoryPoint*)(trj2->GetPoint(0)))->GetPosition();
	    G4double dist = (vert1-vert2).mag();
	    if (m_verboseLevel > 2) G4cout 
	      << "[GateTrajectoryNavigator::FindAnnihilationGammasTrackID] : distance between gammas vertices : dist (mm) " << dist/mm << G4endl;
	    if (dist/mm < 1E-7) {
	      if (m_verboseLevel > 1) {
		G4cout << "[GateTrajectoryNavigator::FindAnnihilationGammasTrackID] : Found common vertex for the two annihilation gammas :" 
		       << " tracks " << trj1->GetTrackID() << " and " << trj2->GetTrackID() << G4endl;
	      }
	      // we add both photons to the vertex
	      m_photonIDVec.push_back(trj1->GetTrackID());
	      m_photonIDVec.push_back(trj2->GetTrackID());
	      // we cancel the 2nd photon from the list to avoid double counting later
	      photonIndices[j2] = -1;
	    }
	  }
	}
      }
    }
  } 
  if ((m_positronTrackID != 0) && ((m_photonIDVec.size() > 2) || (m_photonIDVec.size() == 1))) {
    // strange case: we would expect to see a true annihilation, thus only 2 gammas backto back from the positron
    // 0 can be normal (positron escaped and didn't annihilate), 1 is strange again.
    G4cout 
      << "[GateTrajectoryNavigator::FindAnnihilationGammasTrackID]  WARNING : positron found and photonID vector size: " << m_photonIDVec.size() << G4endl;
  } 
  if ((m_positronTrackID != 0) && (m_photonIDVec.size() == 0)) {
    // still, it's not really what we want to see the positron going out of the scene without doing anything...
    //    if (m_verboseLevel > 0) 
    if (m_verboseLevel > 0)
      G4cout
       << "[GateTrajectoryNavigator::FindAnnihilationGammasTrackID]  WARNING : positron found and photonID vector size: " << m_photonIDVec.size() << G4endl;
    G4Navigator *gNavigator = G4TransportationManager::GetTransportationManager()->GetNavigatorForTracking();
    G4ThreeVector null(0.,0.,0.);
    G4ThreeVector *ptr;
    ptr = &null;
    G4ThreeVector lastPoint = ((G4TrajectoryPoint*)(m_positronTrj->GetPoint(m_positronTrj->GetPointEntries()-1)))->GetPosition();
    G4String theLastVolumeName = gNavigator->LocateGlobalPointAndSetup(lastPoint,ptr,false)->GetName();
    if (m_verboseLevel > 0)
      G4cout << "GateTrajectoryNavigator::FindAnnihilationGammasTrackID: last position of the positron track has been found in volume: " << theLastVolumeName << G4endl;  
  } 
  return m_photonIDVec;
}


// search the gamma from which this track comes --> photonID
G4int GateTrajectoryNavigator::FindPhotonID(G4int trackID)
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::FindPhotonID " << G4endl;
  G4int photonID = 0;

  if (!m_trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::FindPhotonID : ERROR : NULL trajectoryContainer" << G4endl;
  } else {

    if (m_photonIDVec.size() == 0) {
      G4cout << "GateTrajectoryNavigator::FindPhotonID : m_photonIDVec.size() == 0" << G4endl;
    } else {

      G4bool found;
      // search the gamma related to this hit --> photonID
      photonID = trackID;
      G4int photon1ID = m_photonIDVec[0];
      // in case there are 2 gammas OK, if not we put 0 (compatible with subsequent analysis)
      G4int photon2ID = (m_photonIDVec.size() >= 2) ? m_photonIDVec[1] : 0;
      G4int rootID = 0;

      // we go up and up, starting from the present trackID, to the parentID, the parentID, ecc until
      // we find that the ID of the track is equal to the ID of: one of the photons, or rootID(==0)
      while (!((photonID==photon1ID)||(photonID==photon2ID)||(photonID==rootID))) {
	found = false;
	G4int n_trajectories = m_trajectoryContainer->entries();
	for (G4int iTrj=0; (iTrj<n_trajectories) && (!found); iTrj++) {
	  G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
	  if (photonID == trj->GetTrackID()) {
	    photonID = trj->GetParentID();
	    found = true;
	  }
	}
      }
      if (photonID == rootID) {
	if (m_verboseLevel > 2) G4cout
	  << "GateTrajectoryNavigator::FindPhotonID : trackID: " << trackID << " photonID = " << rootID << G4endl; 
      }
      if (photonID == photon1ID) {
	photonID = 1;
      } else if (photonID == photon2ID) {
	photonID = 2;
      }
      
    }
  }
  return photonID;
}


G4int GateTrajectoryNavigator::FindPrimaryID(G4int trackID)
{
  G4int primaryID = 0;

  if (!m_trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::FindPrimaryID : ERROR : NULL trajectoryContainer" << G4endl;
  } else {
    G4int tempParentID = trackID;
    G4int primaryIndex = -1;
    G4bool found;
    G4int n_trajectories = m_trajectoryContainer->entries();
    // we go up and up starting from the trackID, via the parentID's, until
    // the track we find is a primary (its parentID==0)
    do {
      primaryID = tempParentID;
      found = false;
      for (G4int iTrj=0; (iTrj<n_trajectories) && (!found); iTrj++) {
	G4Trajectory* trj = (G4Trajectory*)((*m_trajectoryContainer)[iTrj]);
	if (primaryID == trj->GetTrackID()) {
	  tempParentID = trj->GetParentID();
	  primaryIndex = iTrj;
	  found = true;
	}
      }
    } while (tempParentID != 0);
  }

  return primaryID;
}


void GateTrajectoryNavigator::SetTrajectoryContainer(G4TrajectoryContainer* trajectoryContainer)
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::SetTrajectoryContainer" << G4endl;

  if (!trajectoryContainer) {
    G4cout << "GateTrajectoryNavigator::SetTrajectoryContainer : ERROR : NULL trajectoryContainer" << G4endl;
  } else {
    m_trajectoryContainer = trajectoryContainer;
  }

  Initialize();
}

void GateTrajectoryNavigator::Initialize()
{
  if (m_verboseLevel > 2)
    G4cout << "GateTrajectoryNavigator::Initialize" << G4endl;
  m_photonIDVec.clear();
  m_positronTrackID = -1;
}
